@@ -15,7 +15,6 @@ class Pipeline(object):
 
     def __init__(self, stages: List[Stage]):
         self.stages = stages
-        self.terminated = False
 
 
 class SyncPipeline(Pipeline):
@@ -25,7 +24,7 @@ class SyncPipeline(Pipeline):
     All stages run in the main process.
     '''
 
-    def __init__(self,  stages: List[Stage], queue_size: int = 8):
+    def __init__(self,  stages: List[Stage], queue_size: int = 16):
         super(SyncPipeline, self).__init__(stages)
 
     def start(self, timeout: Optional[int] = 1):
@@ -68,7 +67,7 @@ class AsyncPipeline(Pipeline):
     Asynchronized pipeline powered by multiprocessing.
     '''
 
-    def __init__(self, stages: List[Stage], queue_size: Optional[int] = 8):
+    def __init__(self, stages: List[Stage], queue_size: Optional[int] = 16):
         super(AsyncPipeline, self).__init__(stages)
         self.manager = mp.Manager()
         self.job_queue = self.manager.Queue()
@@ -80,6 +79,8 @@ class AsyncPipeline(Pipeline):
             self.worker_groups.append(group)
             job_queue = group.result_queue
         self.result_queue = job_queue
+        self.resetting = False
+        self.terminated = False
 
     def start(self, timeout: Optional[int] = 1):
         for group in self.worker_groups:
@@ -87,19 +88,25 @@ class AsyncPipeline(Pipeline):
         self.reset(timeout)
 
     def reset(self, timeout: Optional[int] = 1):
+        assert not self.resetting, 'Cannot reset when reset is in progress'
         reset_task = ControlTask(ControlCommand.Reset)
         for _ in range(self.stages[0].num_worker):
             self.job_queue.put(reset_task, timeout=timeout)
+        self.resetting = True
 
     def wait(self, timeout: Optional[int] = None):
         while True:
             result = self.result_queue.get(timeout=timeout)
             if isinstance(result, ControlTask):
+                if result.command == ControlCommand.Reset:
+                    self.resetting = False
                 break
             else:
                 yield result
 
     def end(self, timeout: Optional[int] = 1):
+        if self.terminated:
+            return
         end_task = ControlTask(ControlCommand.End)
         for _ in range(self.stages[0].num_worker):
             self.job_queue.put(end_task, timeout=timeout)
@@ -110,7 +117,7 @@ class AsyncPipeline(Pipeline):
         for group in self.worker_groups:
             group.join(timeout)
 
-    def terminate(self, timeout: Optional[int] = 1):
+    def terminate(self, timeout: Optional[int] = 3):
         for group in self.worker_groups:
             group.terminate(timeout)
         self.terminated = True
