@@ -1,5 +1,6 @@
 from functools import partial
 from queue import SimpleQueue
+import queue
 from threading import Event, Thread
 from typing import Any, Iterable, List, Optional
 
@@ -22,16 +23,17 @@ class Job(object):
         self.name = name
         self.task = task
         self.length = length
-        self.max_retry = max_retry
-        self.results = None
+        self.retry = max_retry
+        self.success = False
 
     def finish(self, results: Any, pipeline_id: int):
         self.results = results
         self.pipeline_id = pipeline_id
+        self.success = True
 
     def __repr__(self):
-        return '%s(%s, len=%d)' % (
-            self.__class__.__name__, self.name, self.length)
+        return '%s(%s, len=%d, retry=%d)' % (
+            self.__class__.__name__, self.name, self.length, self.retry)
 
 
 class System(object):
@@ -40,13 +42,9 @@ class System(object):
     A system consists of multiple peer pipelines.
     '''
 
-    def __init__(self, *, pipeline_queue_size: int = 8,
-                 task_timeout: Optional[int] = 10,
-                 queue_timeout: Optional[int] = 1,
+    def __init__(self, *, pipeline_queue_size: int = 16,
                  resource_scan_args={}, **custom_args):
         self.pipeline_queue_size = pipeline_queue_size
-        self.task_timeout = task_timeout
-        self.queue_timeout = queue_timeout
         self.debug_mode = Options.single_sync_pipeline
         self.resources = Resources(**resource_scan_args)
         self.num_pipeline = self.get_num_pipeline(
@@ -87,27 +85,35 @@ class System(object):
         results = [task.content for task in results_gen]
         return results
 
-    def monit_pipeline(self, pipeline_id: int):
+    def monit_pipeline(self, pipeline_id: int, *,
+                       task_timeout: Optional[int] = None,
+                       queue_timeout: Optional[int] = None):
+        logger = get_logger(
+            '%s(%d)' % (self.pipeline_cls.__name__, pipeline_id))
+        job, job_ok = None, True
+        pipeline_ok = True
         try:
             while True:
                 job = self.job_queue.get()
                 if job is None:
                     return
+                job_ok = False
                 pipeline = self.pipelines[pipeline_id]
-                if pipeline.terminated:
-                    self.logger.info('Pipeline %d: restarting', pipeline_id)
+                if not pipeline_ok:
+                    logger.info('Restarting')
                     pipeline = self.pipeline_fns[pipeline_id]()
-                    pipeline.start()
+                    pipeline.start(queue_timeout)
+                    [*pipeline.wait(task_timeout)]
                     self.pipelines[pipeline_id] = pipeline
+                    logger.info('Restarted')
+                pipeline_ok = False
                 try:
                     if Options.no_progress_bar:
-                        self.logger.info('Pipeline %d: processing job: %s',
-                                         pipeline_id, job.name)
+                        logger.info('Processing: %s', job)
                     if not self.debug_mode:
-                        pipeline.job_queue.put(
-                            job.task, timeout=self.task_timeout)
-                        pipeline.reset(self.queue_timeout)
-                        results_gen = pipeline.wait(self.task_timeout)
+                        pipeline.job_queue.put(job.task, timeout=queue_timeout)
+                        pipeline.reset(queue_timeout)
+                        results_gen = pipeline.wait(task_timeout)
                     else:
                         results_gen = pipeline.run_task(job.task)
                     results_gen = progressbar(
@@ -117,32 +123,31 @@ class System(object):
                         silent=Options.no_progress_bar)
                     results = self.get_results(job, results_gen)
                     if self.debug_mode:
-                        pipeline.reset(self.queue_timeout)
+                        pipeline.reset(queue_timeout)
+                    pipeline_ok = True
+                    if results is None:
+                        raise ValueError('Empty results from %s' % (job))
                     job.finish(results, pipeline_id)
                     if Options.no_progress_bar:
-                        self.logger.info('Pipeline %d: processed job: %s',
-                                         job.pipeline_id, job.name)
+                        logger.info('Processed: %s', job)
                 except Exception as e:
                     if self.ending.is_set():
                         return
-                    self.logger.exception(
-                        'Pipeline %d: failed job: %s', pipeline_id, job.name)
-                    if job.max_retry > 0:
-                        job.max_retry -= 1
-                        self.logger.info(
-                            'Pipeline %d: %s will retry (%d chances left)',
-                            pipeline_id, job, job.max_retry)
-                        self.job_queue.put(job, timeout=self.queue_timeout)
-                    else:
-                        self.result_queue.put(job, timeout=self.queue_timeout)
-                    pipeline.terminate(self.queue_timeout)
+                    logger.exception('Failed: %s', job)
                     if Options.raise_exception:
                         raise e
-                    continue
-                self.result_queue.put(job, timeout=self.queue_timeout)
+                    if not pipeline_ok:
+                        logger.info('Terminating')
+                        pipeline.terminate(queue_timeout * 3)
+                finally:
+                    self.result_queue.put(job, timeout=queue_timeout)
+                    job_ok = True
         except Exception as e:
-            self.logger.exception('Pipeline %d: dead', pipeline_id)
-            self.result_queue.put(None, timeout=self.task_timeout)
+            logger.exception('Dead')
+            if not job_ok:
+                self.result_queue.put(job, timeout=queue_timeout)
+                job_ok = True
+            self.result_queue.put(None, timeout=queue_timeout)
             if Options.raise_exception:
                 raise e
 
@@ -158,22 +163,24 @@ class System(object):
             self.pipeline_fns.append(pipeline_fn)
             pipeline = pipeline_fn()
             self.pipelines.append(pipeline)
-        if not self.debug_mode:
-            self.monitor_threads = [
-                Thread(target=self.monit_pipeline, args=(i,), daemon=True)
-                for i in range(self.num_pipeline)]
         self.progressbar = None
 
-    def start(self, *, run_timeout: Optional[int] = 60,
-              cmd_timeout: Optional[int] = 1):
+    def start(self, *, task_timeout: Optional[int] = 600,
+              queue_timeout: Optional[int] = 1):
         self.logger.info('Starting')
         for pipeline in self.pipelines:
-            pipeline.start(cmd_timeout)
+            pipeline.start(queue_timeout)
         if not self.debug_mode:
-            for thread in self.monitor_threads:
+            self.monitor_threads = []
+            for i in range(self.num_pipeline):
+                thread = Thread(
+                    target=self.monit_pipeline, args=(i,),
+                    kwargs=dict(task_timeout=task_timeout,
+                                queue_timeout=queue_timeout), daemon=True)
                 thread.start()
+                self.monitor_threads.append(thread)
             for pipeline in self.pipelines:
-                [*pipeline.wait(run_timeout)]
+                [*pipeline.wait(task_timeout)]
         self.logger.info('Started')
         if not Options.no_progress_bar:
             self.progressbar = progressbar(
@@ -187,21 +194,30 @@ class System(object):
             self.progressbar.refresh()
         if self.debug_mode:
             self.job_queue.put(None, timeout=timeout)
-            self.monit_pipeline(0)
 
     def add_jobs(self, jobs: List[Job], *, timeout: Optional[int] = 1):
         for job in jobs:
             self.add_job(job, timeout=timeout)
 
-    def wait_job(self, *, timeout: Optional[int] = 1200):
+    def wait_job(self, *, job_timeout: Optional[int] = 3600,
+                 queue_timeout: Optional[int] = 1):
         job = None
-        for _ in range(self.num_pipeline):
-            job = self.result_queue.get(timeout=timeout)
-            if job is None:
+        if self.debug_mode:
+            self.monit_pipeline(0)
+        while self.num_pipeline > 0:
+            job = self.result_queue.get(timeout=job_timeout)
+            if job is None and not self.debug_mode:
                 self.num_pipeline -= 1
-                if self.num_pipeline == 0:
-                    raise ValueError('All pipelines dead')
+            if not job.success:
+                if job.retry > 0:
+                    job.retry -= 1
+                    self.logger.info('Enqueue for retry: %s', job)
+                    self.job_queue.put(job, timeout=queue_timeout)
+                    continue
+                self.logger.warn('Failed: %s', job)
             break
+        else:
+            raise ValueError('All pipelines dead')
         self.result_count += 1
         if not Options.no_progress_bar:
             self.progressbar.update()
@@ -210,12 +226,16 @@ class System(object):
                 self.result_count, self.job_count))
         return job
 
-    def wait_jobs(self, num_jobs: int, *, timeout: Optional[int] = 1200):
+    def wait_jobs(self, num_jobs: int, *, job_timeout: Optional[int] = 3600,
+                  queue_timeout: Optional[int] = 1):
         for _ in range(num_jobs):
-            yield self.wait_job(timeout=timeout)
+            yield self.wait_job(
+                job_timeout=job_timeout, queue_timeout=queue_timeout)
 
     def end(self, *, timeout: Optional[int] = 1):
         self.logger.info('Ending')
+        assert self.job_queue.empty(), 'Ending with unprocessed jobs'
+        assert self.result_queue.empty(), 'Ending with retrieved results'
         self.ending.set()
         for pipeline in self.pipelines:
             pipeline.end(timeout)
@@ -230,7 +250,7 @@ class System(object):
             self.progressbar.close()
         self.logger.info('Ended')
 
-    def terminate(self, *, timeout: Optional[int] = 1):
+    def terminate(self, *, timeout: Optional[int] = 3):
         self.logger.exception('Terminating')
         self.ending.set()
         for pipeline in self.pipelines:
